(*
 * Builtin objects.
 *
 * ----------------------------------------------------------------
 *
 * @begin[license]
 * Copyright (C) 2004 Mojave Group, Caltech
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Author: Jason Hickey
 * @email{jyh@cs.caltech.edu}
 * @end[license]
 *)
open Lm_printf

open Lm_symbol
open Lm_location
open Lm_string_set

open Omake_ir
open Omake_env
open Omake_eval
open Omake_wild
open Omake_node
open Omake_exec
open Omake_value
open Omake_state
open Omake_symbol
open Omake_cache_type
open Omake_builtin_fun
open Omake_builtin_util

module Pos = MakePos (struct let name = "Omake_builtin_object" end)
open Pos

(*
 * Extend an object with another.
 * The argument may be a file or an object.
 *)
let object_of_file venv pos loc s =
   let pos  = string_pos "extends" pos in
   let cache = venv_cache venv in
   let rec search l =
      match l with
         v :: l ->
            let dir = dir_of_value venv pos v in
            let node = Node.intern dir s in
               if Omake_cache.exists cache node then
                  node
               else
                  search l
       | [] ->
            raise (OmakeException (loc_pos loc pos, StringStringError ("can't find", s)))
   in
   let path = values_of_value venv pos (venv_find_var venv ScopePublic pos loc omakepath_sym) in
   let node = search path in
      eval_object_file venv pos loc node

(*
 * Extend an object with another.
 * The argument may be a file or an object.
 *)
let extends_fun venv pos loc args =
   let pos = string_pos "extends" pos in
   let extend_val venv v =
      let obj =
         match eval_value venv pos v with
            ValObject obj ->
               obj
          | v ->
               object_of_file venv pos loc (string_of_value venv pos v)
      in
         venv_include_object venv obj
   in
   let extend_arg venv arg =
      let values = values_of_value venv pos arg in
         List.fold_left extend_val venv values
   in
   let venv = List.fold_left extend_arg venv args in
      ValEnv (venv, ExportAll)

(*
 * Import an object from a file.
 * And add that object by its name to the environment.
 *
 * The argument may be a file or an object,
 * but if it is an object, this does nothing.
 *)
let import_fun venv pos loc args =
   let pos = string_pos "import" pos in
   let import_val venv v =
      match eval_value venv pos v with
         ValObject _ ->
            venv
       | _ ->
            let name = string_of_value venv pos v in
            let obj = object_of_file venv pos loc name in
            let name = Filename.basename name in
            let name = Lm_filename_util.root name in
            let name = Lm_symbol.add name in
               venv_add_var venv ScopePublic pos name (ValObject obj)
   in
   let import_arg venv arg =
      let values = values_of_value venv pos arg in
         List.fold_left import_val venv values
   in
   let venv = List.fold_left import_arg venv args in
      ValEnv (venv, ExportAll)

(*
 * Normal objects.
 *)
let eval_obj venv pos v =
   let obj = eval_object venv pos v in
      obj, obj

let wrap_obj _ obj =
   ValObject obj

(*
 * For getting map objects.
 *)
let eval_map venv pos v =
   let pos = string_pos "eval_map" pos in
   let obj = eval_object venv pos v in
      match eval_object_value venv pos obj with
         ValMap map ->
            obj, map
       | _ ->
            raise (OmakeException (pos, StringError "not a map"))

let wrap_map obj map =
   ValObject (add_object_value obj (ValMap map))

(*
 * Field membership.
 *)
let object_mem eval_object venv pos loc args =
   let pos = string_pos "object-mem" pos in
      match args with
         [arg; v] ->
            let _, obj = eval_object venv pos arg in
            let s = string_of_value venv pos v in
            let v = Lm_symbol.add s in
               if venv_object_mem obj v then
                  val_true
               else
                  val_false
       | _ ->
            raise (OmakeException (loc_pos loc pos, ArityMismatch (ArityExact 2, List.length args)))

let object_find eval_object venv pos loc args =
   let pos = string_pos "object-find" pos in
      match args with
         [arg; v] ->
            let _, obj = eval_object venv pos arg in
            let s = string_of_value venv pos v in
            let v = Lm_symbol.add s in
               venv_find_field obj pos v
       | _ ->
            raise (OmakeException (loc_pos loc pos, ArityMismatch (ArityExact 2, List.length args)))

(*
 * Add a field to an object.
 *)
let object_add eval_obj wrap_obj venv pos loc args =
   let pos = string_pos "object-add" pos in
      match args with
         [arg; v; x] ->
            let root, obj = eval_obj venv pos arg in
            let s = string_of_value venv pos v in
            let v = Lm_symbol.add s in
               wrap_obj root (venv_add_field obj v x)
       | _ ->
            raise (OmakeException (loc_pos loc pos, ArityMismatch (ArityExact 2, List.length args)))

(*
 * Add a field to an object.
 *)
let object_length eval_object venv pos loc args =
   let pos = string_pos "object-length" pos in
      match args with
         [arg] ->
            let _, obj = eval_object venv pos arg in
               ValInt (venv_object_length obj)
       | _ ->
            raise (OmakeException (loc_pos loc pos, ArityMismatch (ArityExact 1, List.length args)))

(*
 * Iterate over the object.
 *)
let object_map eval_obj wrap_obj venv pos loc args =
   let pos = string_pos "map" pos in
   let f, root, obj =
      match args with
         [arg; fun_val] ->
            let root, obj = eval_obj venv pos arg in
            let _, _, f = eval_fun venv pos fun_val in
               f, root, obj
       | [arg; param1; param2; ValBody body] ->
            let params =
               [Lm_symbol.add (string_of_value venv pos param1);
                Lm_symbol.add (string_of_value venv pos param2)]
            in
            let root, obj = eval_obj venv pos arg in
            let env = venv_get_env venv in
            let f venv pos loc args =
               let venv = venv_add_args venv pos loc env params args in
                  eval venv body
            in
               f, root, obj
       | _ ->
            raise (OmakeException (loc_pos loc pos, ArityMismatch (ArityRange (3, 4), List.length args)))
   in

   (* If the body exports the environment, preserve it across calls *)
   let venv', obj =
      venv_object_fold (fun (venv, obj) v x ->
            let result = f venv pos loc [ValString (Lm_symbol.to_string v); x] in
            let obj = venv_add_field obj v result in
            let venv, result = add_exports venv pos result in
               venv, obj) (venv, obj) obj
   in
      if venv' == venv then
         wrap_obj root obj
      else
         ValEnv (venv', ExportAll)

(*
 * Return the number of elements in the sequence.
 *)
let sequence_length venv pos loc args =
   let pos = string_pos "length" pos in
      match args with
         [arg] ->
            let obj = eval_object venv pos arg in
            let arg = eval_object_value venv pos obj in
            let len =
               match arg with
                  ValMap obj
                | ValObject obj ->
                     venv_object_length obj
                | ValNone
                | ValEnv _ ->
                     0
                | ValInt _
                | ValFloat _
                | ValNode _
                | ValDir _
                | ValBody _
                | ValChannel _ ->
                     1
                | ValData s ->
                     String.length s
                | ValQuote vl ->
                     String.length (string_of_quote venv pos vl)
                | ValSequence _
                | ValString _ ->
                     List.length (values_of_value venv pos arg)
                | ValArray a ->
                     List.length a
                | ValFun (arity, _, _, _)
                | ValFunValue (arity, _, _, _)
                | ValPrim (arity, _, _) ->
                     (match arity with
                         ArityExact i
                       | ArityRange (i, _) ->
                            i
                       | ArityNone ->
                            0
                       | ArityAny ->
                            max_int)
                | ValRules l ->
                     List.length l
                | ValCases cases ->
                     List.length cases
                | ValClass _
                | ValOther _ ->
                     0
                | ValApply _
                | ValSuperApply _
                | ValMethodApply _ ->
                     raise (Invalid_argument "Omake_builtin_base.length")
            in
               ValInt len
       | _ ->
            raise (OmakeException (loc_pos loc pos, ArityMismatch (ArityExact 1, List.length args)))

(*
 * Get the nth element of a sequence.
 *)
let sequence_nth venv pos loc args =
   let pos = string_pos "nth" pos in
      match args with
         [arg; i] ->
            let i = int_of_value venv pos i in
            let obj = eval_object venv pos arg in
            let arg = eval_object_value venv pos obj in
               (match arg with
                   ValNone
                 | ValFun _
                 | ValFunValue _
                 | ValPrim _
                 | ValApply _
                 | ValSuperApply _
                 | ValMethodApply _
                 | ValEnv _
                 | ValMap _
                 | ValObject _ ->
                      raise (OmakeException (loc_pos loc pos, StringIntError ("out of bounds", i)))
                 | ValInt _
                 | ValFloat _
                 | ValNode _
                 | ValDir _
                 | ValBody _
                 | ValChannel _
                 | ValClass _
                 | ValCases _
                 | ValOther _ ->
                      if i = 0 then
                         arg
                      else
                         raise (OmakeException (loc_pos loc pos, StringIntError ("out of bounds", i)))
                 | ValData s ->
                      let len = String.length s in
                         if i < 0 || i >= len then
                            raise (OmakeException (loc_pos loc pos, StringIntError ("out of bounds", i)));
                         ValData (String.sub s i 1)
                 | ValQuote vl ->
                      let s = string_of_quote venv pos vl in
                      let len = String.length s in
                         if i < 0 || i >= len then
                            raise (OmakeException (loc_pos loc pos, StringIntError ("out of bounds", i)));
                         ValData (String.sub s i 1)

                 | ValSequence _
                 | ValString _ ->
                      let values = values_of_value venv pos arg in
                      let len = List.length values in
                         if i < 0 || i >= len then
                            raise (OmakeException (loc_pos loc pos, StringIntError ("out of bounds", i)));
                         List.nth values i
                 | ValArray a ->
                      let len = List.length a in
                         if i < 0 || i >= len then
                            raise (OmakeException (loc_pos loc pos, StringIntError ("out of bounds", i)));
                         List.nth a i
                 | ValRules l ->
                      let len = List.length l in
                         if i < 0 || i >= len then
                            raise (OmakeException (loc_pos loc pos, StringIntError ("out of bounds", i)));
                         ValRules [List.nth l i])
       | _ ->
            raise (OmakeException (loc_pos loc pos, ArityMismatch (ArityExact 2, List.length args)))

(*
 * Reverse the elements in the sequence.
 *)
let sequence_rev venv pos loc args =
   let pos = string_pos "rev" pos in
      match args with
         [arg] ->
            let obj = eval_object venv pos arg in
            let arg = eval_object_value venv pos obj in
               (match arg with
                   ValNone
                 | ValFun _
                 | ValFunValue _
                 | ValPrim _
                 | ValApply _
                 | ValSuperApply _
                 | ValMethodApply _
                 | ValEnv _
                 | ValMap _
                 | ValObject _
                 | ValInt _
                 | ValFloat _
                 | ValNode _
                 | ValDir _
                 | ValBody _
                 | ValChannel _
                 | ValClass _
                 | ValOther _ ->
                      arg
                 | ValData s1 ->
                      let len = String.length s1 in
                      let s2 = String.create len in
                         for i = 0 to len - 1 do
                            s2.[i] <- s1.[len - i - 1]
                         done;
                         ValData s2
                 | ValQuote vl ->
                      let s1 = string_of_quote venv pos vl in
                      let len = String.length s1 in
                      let s2 = String.create len in
                         for i = 0 to len - 1 do
                            s2.[i] <- s1.[len - i - 1]
                         done;
                         ValData s2
                 | ValCases cases ->
                      ValCases (List.rev cases)
                 | ValSequence _
                 | ValString _ ->
                      let values = values_of_value venv pos arg in
                         ValArray (List.rev values)
                 | ValArray a ->
                      ValArray (List.rev a)
                 | ValRules l ->
                      ValRules (List.rev l))
       | _ ->
            raise (OmakeException (loc_pos loc pos, ArityMismatch (ArityExact 0, List.length args)))

(*
 * Map.
 *
 * \begin{doc}
 * \subsection{foreach}
 *
 * The \verb+foreach+ function maps a function over a sequence.
 *
 * \begin{verbatim}
 * $(foreach <fun>, <args>)
 *
 * foreach(<var>, <args>)
 *    <body>
 * \end{verbatim}
 *
 * For example, the following program defines the variable \verb+X+
 * as \verb+a.c b.c c.c+.
 *
 * \begin{verbatim}
 * X =
 *    foreach(x, a b c)
 *       $(x).c
 *
 * # Equivalent expression
 * X = $(foreach $(fun x, $(x).c), abc)
 * \end{verbatim}
 *
 * There is also an abbreviated syntax.
 *
 * The \verb+export+ form can also be used in a \verb+foreach+
 * body.  The final value of \verb+X+ is \verb+a.c b.c c.c+.
 *
 * \begin{verbatim}
 * X =
 * foreach(x, a b c)
 *    X += $(x).c
 *    export
 * \end{verbatim}
 * \end{doc}
 *)
let foreach_fun venv pos loc args =
   let pos = string_pos "foreach" pos in
   let f, args =
      match args with
         [fun_val; arg] ->
            let args = values_of_value venv pos arg in
            let _, _, f = eval_fun venv pos fun_val in
               f, args
       | [ValBody body; param; arg] ->
            let params = [Lm_symbol.add (string_of_value venv pos param)] in
            let args = values_of_value venv pos arg in
            let env = venv_get_env venv in
            let f venv pos loc args =
               let venv = venv_add_args venv pos loc env params args in
                  eval venv body
            in
               f, args
       | _ ->
            raise (OmakeException (loc_pos loc pos, ArityMismatch (ArityRange (2, 3), List.length args)))
   in

   (* If the body exports the environment, preserve it across calls *)
   let venv', values =
      List.fold_left (fun (venv, values) v ->
            let result = f venv pos loc [v] in
            let venv, result = add_exports venv pos result in
               venv, result :: values) (venv, []) args
   in
      if venv' == venv then
         ValArray (List.rev values)
      else
         ValEnv (venv', ExportAll)

(*
 * Add only the builtin functions.
 * The Pervasives file defines most of the remaining methods.
 *)
let builtin_funs =
   [true, "extends",              extends_fun,                                ArityAny;
    true, "import",               import_fun,                                 ArityAny;
    true, "foreach",              foreach_fun,                                ArityRange (2, 3);
    true, "obj-add",              object_add eval_obj wrap_obj,               ArityExact 3;
    true, "obj-find",             object_find eval_obj,                       ArityExact 2;
    true, "obj-mem",              object_mem eval_obj,                        ArityExact 2;
    true, "obj-length",           object_length eval_obj,                     ArityExact 1;
    true, "obj-map",              object_map eval_obj wrap_obj,               ArityRange (3, 4);
    true, "map-add",              object_add eval_map wrap_map,               ArityExact 3;
    true, "map-find",             object_find eval_map,                       ArityExact 2;
    true, "map-mem",              object_mem eval_map,                        ArityExact 2;
    true, "map-length",           object_length eval_map,                     ArityExact 1;
    true, "map-map",              object_map eval_map wrap_map,               ArityRange (3, 4);
    true, "sequence-foreach",     foreach_fun,                                ArityRange (2, 3);
    true, "sequence-length",      sequence_length,                            ArityExact 1;
    true, "sequence-nth",         sequence_nth,                               ArityExact 1;
    true, "sequence-rev",         sequence_rev,                               ArityExact 1]

let builtin_objects =
   ["Int",              ValInt 0;
    "Float",            ValFloat 0.0;
    "String",           ValNone;
    "Array",            ValArray [];
    "Fun",              ValFunValue (ArityExact 0, venv_empty_env, [], ValNone);
    "Rule",             ValRules [];
    "File",             ValNone;
    "Dir",              ValNone;
    "Body",             ValNone;
    "InChannel",        ValNone;
    "OutChannel",       ValNone;
    "InOutChannel",     ValNone;
    "Map",              ValMap venv_empty_object]

let pervasive_objects =
   ["Object";
    "Number";
    "Sequence";
    "Node";
    "Channel";
    "Exception";
    "RuntimeException";
    "Select";
    "Pipe";
    "Stat";
    "Shell";
    "Lexer";
    "Parser";
    "Location";
    "Position"]

let builtin_rules = []
let builtin_vars = []
let phony_targets = []

(*!
 * @docoff
 *
 * -*-
 * Local Variables:
 * Caml-master: "compile"
 * End:
 * -*-
 *)
