(*
 * Predefined set of functions.
 *
 * ----------------------------------------------------------------
 *
 * @begin[license]
 * Copyright (C) 2003 Jason Hickey, Caltech
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Author: Jason Hickey
 * @email{jyh@cs.caltech.edu}
 * @end[license]
 *)
open Lm_glob
open Lm_debug
open Lm_printf
open Lm_symbol
open Lm_location
open Lm_string_set

open Omake_ir
open Omake_env
open Omake_exec
open Omake_wild
open Omake_node
open Omake_util
open Omake_state
open Omake_ir_ast
open Omake_command
open Omake_ir_print
open Omake_node_sig
open Omake_exec_type
open Omake_exec_util
open Omake_cache_type
open Omake_shell_type
open Omake_shell_job
open Omake_symbol

module Pos = MakePos (struct let name = "Omake_eval" end)
open Pos

let debug_eval =
   create_debug (**)
      { debug_name = "debug-eval";
        debug_description = "Debug the evaluator";
        debug_value = false
      }

let print_ast =
   create_debug (**)
      { debug_name = "print-ast";
        debug_description = "Print the AST after parsing";
        debug_value = false
      }

let print_ir =
   create_debug (**)
      { debug_name = "print-ir";
        debug_description = "Print the IR after parsing";
        debug_value = false
      }

let print_rules =
    create_debug (**)
      { debug_name = "print-rules";
        debug_description = "Print the rules after evaluation";
        debug_value = false
      }

let print_files =
   create_debug (**)
      { debug_name = "print-files";
        debug_description = "Print the files as they are read";
        debug_value = false
      }

(*
 * From Omake_cache.
 *)
let include_fun = Omake_cache.include_fun
let include_scanner_fun = Omake_cache.include_scanner_fun

(************************************************************************
 * Utilities.
 *)

(*
 * Check whether a value has an embedded array.
 *)
let rec is_array_value v =
   match v with
      ValArray _ ->
         true
    | ValSequence vl ->
         List.exists is_array_value vl
    | ValObject obj ->
         (try
             match venv_find_field_exn obj builtin_sym with
                ValArray _ ->
                   true
              | _ ->
                   false
          with
             Not_found ->
                false)
    | ValNone
    | ValInt _
    | ValFloat _
    | ValData _
    | ValQuote _
    | ValString _
    | ValApply _
    | ValFun _
    | ValFunValue _
    | ValPrim _
    | ValRules _
    | ValNode _
    | ValDir _
    | ValEnv _
    | ValBody _
    | ValMap _
    | ValSuperApply _
    | ValMethodApply _
    | ValChannel _
    | ValClass _
    | ValCases _
    | ValOther _ ->
         false

(*
 * Get a value from a string_exp.
 *)
let strategy_is_eager be_eager strategy =
   match strategy with
      EagerApply ->
         true
    | NormalApply ->
         be_eager
    | LazyApply ->
         false

(************************************************************************
 * Compiling utilities.
 *)

(*
 * The name of an include file.
 *)
let include_name name =
   name ^ ".omc"

let object_name name =
   name ^ ".omo"

(*
 * Parse and evaluate a file.
 *)
let parse_ir node =
   let filename = Node.fullname node in
   let ast = Omake_ast_lex.parse_ast filename in
   let _ =
      if debug print_ast then
         eprintf "@[<v 3>AST:@ %a@]@." Omake_ast_print.pp_print_prog ast
   in
   let senv, ir = Omake_ir_ast.compile_prog senv_empty ast in
   let _ =
      if debug print_ir then
         eprintf "@[<v 3>IR:@ %a@]@." Omake_ir_print.pp_print_exp ir
   in
      senv, ir

(*
 * Marshaling.
 *)
let marshal_from_channel inx magic_number =
   let magic = Omake_magic.input_magic inx in
   let _ =
      if magic <> magic_number then
         raise (Failure "bad magic number")
   in
      Marshal.from_channel inx

let marshal_to_channel venv outx magic_number x =
   Omake_magic.output_magic outx magic_number;
   venv_marshal venv (fun () -> Marshal.to_channel outx x []) ()

(*
 * The include file contains the IR for the file.
 *)
let compile_file output_name compile magic_number venv pos loc source =
   let pos = string_pos "compile_include_file" pos in

   (*
    * Get the name of the result file.
    * If the target directory is not writable, place the file in the .omake directory.
    *)
   let source_name = Node.absname source in
   let dir = Filename.dirname source_name in
   let name = output_name (Filename.basename source_name) in
   let target_name = get_cache_file dir name in
   let target  = Node.intern (Dir.cwd ()) target_name in
   let targets = NodeSet.singleton target in
   let sources = NodeSet.singleton source in

   (* New fetch the saved dependencies *)
   let cache = venv_cache venv in
   let _ =  Omake_cache.force_stat cache source in
   let command = ["Omake_eval.compile_include_file"; "-o"; target_name; source_name] in
   let commands =
      [{ Omake_cache_type.command_loc    = loc;
         Omake_cache_type.command_dir    = venv_dir venv;
         Omake_cache_type.command_target = target;
         Omake_cache_type.command_flags  = [];
         Omake_cache_type.command_venv   = venv;
         Omake_cache_type.command_inst   = CommandArgv command
       }]
   in
   let sources =
      try
         let deps = Omake_cache.find_result cache include_scanner_fun sources commands in
            match deps with
               MemoResult deps ->
                  NodeTable.find deps target
             | MemoNone
             | MemoFailed _ ->
                  sources
      with
         Not_found ->
            sources
   in

   (* Now see if it is already up-to-date *)
   let up_to_date = Omake_cache.up_to_date cache include_fun sources commands in

   (* Compile the file and save it to the target *)
   let save_file () =
      if !print_files then
         eprintf "*** omake: creating %s@." target_name;

      let sources, ir =
         try compile source with
            Unix.Unix_error _
          | Sys_error _ as exn ->
               raise (UncaughtException (pos, exn))
      in
      let sources = NodeSet.add sources source in
      let deps = NodeTable.add NodeTable.empty target sources in
      let () =
         (* Try to unlink the file, in case it is not writable *)
         try Unix.unlink target_name with
            Unix.Unix_error _ ->
               ()
      in
      let () =
         try
            let outx = Pervasives.open_out_bin target_name in
            let () =
               try marshal_to_channel venv outx magic_number ir with
                  exn ->
                     close_out outx;
                     raise exn
            in
               close_out outx;
               if Omake_cache.force_stat cache target = None then
                  eprintf "*** omake: failed to create compiled file: %a@." pp_print_node target;
               Omake_cache.add cache include_scanner_fun targets sources commands (MemoResult deps);
               Omake_cache.add cache include_fun targets sources commands MemoNone
         with
            Sys_error _ ->
               eprintf "*** omake: file is not writable: %a@." pp_print_node target
      in
         ir
   in

   (* Load the file from the target *)
   let load_file () =
      if !print_files then
         eprintf "*** omake: reading %s@." target_name;
      try
         let inx = open_in_bin target_name in
         let ir =
            try marshal_from_channel inx magic_number with
               exn ->
                  close_in inx;
                  raise exn
         in
            close_in inx;
            ir
      with
         Sys_error _ ->
            eprintf "*** omake: file is not readable: %a@." pp_print_node target;
            save_file ()
       | Failure "bad magic number" ->
            save_file ()
   in
      if up_to_date then
         load_file ()
      else
         save_file ()

(************************************************************************
 * Value operations.
 *)

(*
 * Get the string representation of a value.
 * It is now not legal to convert an array to a string.
 * Hopefully we will catch all the internal uses appropriately.
 *)
let rec string_of_value venv pos v =
   let pos = string_pos "string_of_value" pos in
   let scratch_buf = Buffer.create 32 in
   let dir1 = venv_dir venv in
   let rec collect v =
      match eval_prim_value venv pos v with
         (* Values that expand to nothing *)
         ValNone
       | ValFun _
       | ValFunValue _
       | ValPrim _
       | ValRules _
       | ValBody _
       | ValMap _
       | ValObject _
       | ValEnv _
       | ValChannel _
       | ValClass _
       | ValCases _
       | ValOther _
       | ValArray [] ->
            ()
       | ValSequence vl
       | ValQuote vl ->
            List.iter collect vl
       | ValArray [v] ->
            collect v
       | ValArray vl ->
            let print_error buf =
               eprintf "@[<v 3>Array value where string expected:";
               eprintf "@ Use the $(string ...) function if you really want to do this";
               eprintf "@ @[<v 3>The array has length %d:" (List.length vl);
               ignore (List.fold_left (fun index v ->
                             eprintf "@ @[<hv 3>[%d] =@ %a@]" index pp_print_value v;
                             succ index) 0 vl);
               eprintf "@]@]@."
            in
               raise (OmakeException (pos, LazyError print_error))
       | ValInt i ->
            Buffer.add_string scratch_buf (string_of_int i)
       | ValFloat x ->
            Buffer.add_string scratch_buf (string_of_float x)
       | ValData s
       | ValString s ->
            Buffer.add_string scratch_buf s
       | ValDir dir2 ->
            Buffer.add_string scratch_buf (Dir.name dir1 dir2)
       | ValNode node ->
            Buffer.add_string scratch_buf (Node.name dir1 node)
       | ValApply _
       | ValSuperApply _
       | ValMethodApply _ ->
            raise (Invalid_argument "string_of_value")
   in
      collect v;
      Buffer.contents scratch_buf

(*
 * Collect the values in a quotation into a string.
 * Even array values are flattened without warning.
 *)
and string_of_quote venv pos vl =
   let pos = string_pos "string_of_quote" pos in
   let scratch_buf = Buffer.create 32 in
   let dir1 = venv_dir venv in
   let rec collect v =
      match eval_value venv pos v with
         (* Values that expand to nothing *)
         ValNone
       | ValFun _
       | ValFunValue _
       | ValPrim _
       | ValRules _
       | ValBody _
       | ValMap _
       | ValObject _
       | ValEnv _
       | ValChannel _
       | ValClass _
       | ValCases _
       | ValOther _
       | ValArray [] ->
            ()
       | ValSequence vl
       | ValQuote vl ->
            List.iter collect vl
       | ValArray [v] ->
            collect v
       | ValArray vl ->
            collect_array vl
       | ValInt i ->
            Buffer.add_string scratch_buf (string_of_int i)
       | ValFloat x ->
            Buffer.add_string scratch_buf (string_of_float x)
       | ValData s
       | ValString s ->
            Buffer.add_string scratch_buf s
       | ValDir dir2 ->
            Buffer.add_string scratch_buf (Dir.name dir1 dir2)
       | ValNode node ->
            Buffer.add_string scratch_buf (Node.name dir1 node)
       | ValApply _
       | ValSuperApply _
       | ValMethodApply _ ->
            raise (Invalid_argument "string_of_value")
   and collect_array vl =
      match vl with
         [v] ->
            collect v
       | v :: vl ->
            collect v;
            Buffer.add_char scratch_buf ' ';
            collect_array vl
       | [] ->
            ()
   in
      List.iter collect vl;
      Buffer.contents scratch_buf


(*
 * Get a list of values from the value.
 * Array elements are always special, and returned as an element.
 * We divide values into two classes:
 *    The "catenable" values are the values that can be concatenated to
 *    for a string.  These include: string, node, dir, int, float.
 *
 *    Nothing else can be concatenated with a string, and is always preserved
 *    in the value list.
 *)
and values_of_value venv pos v =
   let pos = string_pos "values_of_value" pos in
   let dir1 = venv_dir venv in

   (*
    * Convert a catenable value to a string
    *)
   let unwrap v =
      match v with
         ValData s ->
            s
       | ValInt i ->
            string_of_int i
       | ValFloat x ->
            string_of_float x
       | ValNode node ->
            Node.name dir1 node
       | ValDir dir2 ->
            Dir.name dir1 dir2
       | ValQuote vl ->
            string_of_quote venv pos vl
       | ValNone
       | ValString _
       | ValSequence _
       | ValArray _
       | ValFun _
       | ValPrim _
       | ValRules _
       | ValEnv _
       | ValBody _
       | ValMap _
       | ValObject _
       | ValChannel _
       | ValClass _
       | ValCases _
       | ValOther _
       | ValFunValue _
       | ValApply _
       | ValSuperApply _
       | ValMethodApply _ ->
            raise (Invalid_argument "values_of_value.unwrap")
   in
   let tokens = Lm_string_util.tokens_create (fun s -> ValData s) unwrap in

   (*
    * Array elements are always separate values.
    * The arrays are flattened.
    *)
   let rec collect_array tokens vl vll =
      match vl, vll with
         v :: vl, _ ->
            (match v with
                ValArray el ->
                   collect_array tokens el (vl :: vll)
              | ValSequence [v] ->
                   collect_array tokens (v :: vl) vll
              | v ->
                   collect_array (Lm_string_util.tokens_atomic tokens v) vl vll)
       | [], vl :: vll ->
            collect_array tokens vl vll
       | [], [] ->
            tokens
   in

   (*
    * Collect_string is used when we have seen whitespace
    * in a sequence.  Collect the values into the string buffer,
    * then parse the string into separate tokens.
    *)
   let rec collect tokens vl vll =
      match vl, vll with
         v :: vl, _ ->
            let v = eval_value venv pos v in
               (match v with
                   ValNone ->
                      collect tokens vl vll

                   (* Strings *)
                 | ValString s ->
                      collect (Lm_string_util.tokens_string tokens s) vl vll
                 | ValSequence el ->
                      collect tokens el (vl :: vll)

                   (* Other catenable values *)
                 | ValData _
                 | ValInt _
                 | ValFloat _
                 | ValDir _
                 | ValNode _
                 | ValQuote _ ->
                      collect (Lm_string_util.tokens_add tokens v) vl vll

                   (* Atomic values *)
                 | ValArray el ->
                      collect (collect_array (Lm_string_util.tokens_break tokens) el []) vl vll
                 | ValFun _
                 | ValPrim _
                 | ValRules _
                 | ValEnv _
                 | ValBody _
                 | ValMap _
                 | ValObject _
                 | ValChannel _
                 | ValClass _
                 | ValCases _
                 | ValOther _
                 | ValFunValue _
                 | ValApply _
                 | ValSuperApply _
                 | ValMethodApply _ ->
                      collect (Lm_string_util.tokens_atomic tokens v) vl vll)

       | [], vl :: vll ->
            collect tokens vl vll
       | [], [] ->
            Lm_string_util.tokens_flush tokens
   in
      collect tokens [v] []

(*
 * Get a string list from the value.
 * This is always legal because arrays have been flattened.
 *)
and strings_of_value venv pos v =
   let values = values_of_value venv pos v in
      List.map (string_of_value venv pos) values

(*
 * Boolean test.
 * Arrays are always true.
 *)
and bool_of_value venv pos v =
   let values = values_of_value venv pos v in
      match values with
         [] ->
            false
       | [ValData s]
       | [ValString s] ->
            Omake_util.bool_of_string s
       | [ValQuote vl] ->
            Omake_util.bool_of_string (string_of_quote venv pos vl)
       | _ ->
            true

(*
 * The value should be a directory.
 *)
and file_of_value venv pos file =
   let pos = string_pos "file_of_value" pos in
   let file = eval_prim_value venv pos file in
      match file with
         ValNode node ->
            node
       | ValDir _
       | ValData _
       | ValString _
       | ValSequence _
       | ValQuote _
       | ValInt _
       | ValFloat _ ->
            venv_intern venv (string_of_value venv pos file)
       | ValArray _
       | ValNone
       | ValApply _
       | ValFun _
       | ValFunValue _
       | ValPrim _
       | ValRules _
       | ValEnv _
       | ValBody _
       | ValMap _
       | ValObject _
       | ValSuperApply _
       | ValMethodApply _
       | ValChannel _
       | ValClass _
       | ValCases _
       | ValOther _ ->
            raise (OmakeException (pos, StringError "illegal value"))

(************************************************************************
 * Evaluation.
 *)

(*
 * Unfold the outermost application to get a real value.
 *)
and eval_value venv pos v =
   match v with
      ValApply (loc, scope, v, []) ->
         eval_value venv pos (eval_var venv pos loc (venv_find_var venv scope pos loc v))
    | ValApply (loc, scope, v, args) ->
         eval_value venv pos (eval_apply venv pos loc (venv_find_var venv scope pos loc v) args)
    | ValSuperApply (loc, scope, super, v, args) ->
         let obj = venv_find_super venv pos loc super in
         let v = venv_find_field obj pos v in
            eval_value venv pos (eval_apply venv pos loc v args)
    | ValMethodApply (loc, scope, vl, args) ->
         let obj, v = eval_method_var venv scope pos loc vl in
         let venv = venv_with_object venv obj in
            eval_value venv pos (eval_apply venv pos loc v args)
    | ValSequence [v] ->
         eval_value venv pos v
    | _ ->
         v

and eval_single_value venv pos v =
   let pos = string_pos "eval_single_value" pos in
      match eval_value venv pos v with
         ValArray [v] ->
            eval_single_value venv pos v
       | _ ->
            v

and eval_prim_value venv pos v =
   let pos = string_pos "eval_prim_value" pos in
   let v = eval_value venv pos v in
      match v with
         ValArray [v] ->
            eval_prim_value venv pos v
       | ValObject obj ->
            (try venv_find_field_exn obj builtin_sym with
                Not_found ->
                   v)
       | _ ->
            v

(*
 * Evaluate the value in a function body.
 * Expand all applications.
 *)
and eval_body_value venv pos v =
   match eval_value venv pos v with
      ValSequence sl ->
         ValSequence (List.map (eval_body_value venv pos) sl)
    | ValArray sl ->
         ValArray (List.map (eval_body_value venv pos) sl)
    | ValBody body ->
         let _, result = eval_exp venv ValNone body in
            result
    | ValNone
    | ValInt _
    | ValFloat _
    | ValData _
    | ValString _
    | ValQuote _
    | ValDir _
    | ValNode _
    | ValFun _
    | ValFunValue _
    | ValPrim _
    | ValRules _
    | ValMap _
    | ValObject _
    | ValEnv _
    | ValChannel _
    | ValClass _
    | ValCases _
    | ValOther _ ->
         v
    | ValApply _
    | ValSuperApply _
    | ValMethodApply _ ->
         raise (Invalid_argument "eval_body_value")

and eval_body_exp venv pos x v =
   match eval_value venv pos v with
      ValSequence sl ->
         venv, ValSequence (List.map (eval_body_value venv pos) sl)
    | ValArray sl ->
         venv, ValArray (List.map (eval_body_value venv pos) sl)
    | ValBody body ->
         eval_exp venv x body
    | ValNone
    | ValInt _
    | ValFloat _
    | ValData _
    | ValQuote _
    | ValString _
    | ValDir _
    | ValNode _
    | ValFun _
    | ValFunValue _
    | ValPrim _
    | ValRules _
    | ValMap _
    | ValObject _
    | ValEnv _
    | ValChannel _
    | ValClass _
    | ValCases _
    | ValOther _ ->
         venv, v
    | ValApply _
    | ValSuperApply _
    | ValMethodApply _ ->
         raise (Invalid_argument "eval_body_exp")

(*
 * Evaluate a variable.
 * It is fine for the variable to evaluate to a function.
 * But if the function has arity 0, then evaluate it.
 *)
and eval_var venv pos loc v =
   match eval_value venv pos v with
      ValFun (_, env, [], body) ->
         let venv = venv_with_env venv env in
         let _, result = eval_exp venv ValNone body in
            result
    | ValFunValue (_, env, [], v) ->
         let venv = venv_with_env venv env in
            eval_body_value venv pos v
    | ValPrim (ArityExact 0, _, f) ->
         venv_apply_prim_fun f venv pos loc []
    | _ ->
         v

(*
 * Evaluate an application.
 *)
and eval_apply venv pos loc v args =
   match eval_value venv pos v with
      ValFun (_, env, params, body) ->
         let venv = venv_add_args venv pos loc env params args in
         let _, result = eval_exp venv ValNone body in
            result
    | ValFunValue (_, env, params, body) ->
         let venv = venv_add_args venv pos loc env params args in
            eval_body_value venv pos body
    | ValPrim (_, _, f) ->
         venv_apply_prim_fun f venv pos loc args
    | v ->
         if args = [] then
            v
         else
            let print_error buf =
               fprintf buf "@[<v 3>illegal function application:@ @[<hv 3>function:@ %a@]" pp_print_value v;
               List.iter (fun arg ->
                     fprintf buf "@ @[<hv 3>arg = %a@]" pp_print_value arg) args;
               fprintf buf "@]"
            in
               raise (OmakeException (pos, LazyError print_error))

(*
 * Evaluate an application with string arguments.
 *)
and eval_apply_string_exp venv_args venv pos loc v args =
   match eval_value venv pos v with
      ValFun (_, env, params, body) ->
         let args = List.map (eval_string_exp true venv_args pos) args in
         let venv = venv_add_args venv pos loc env params args in
         let _, result = eval_exp venv ValNone body in
            result
    | ValFunValue (_, env, params, body) ->
         let args = List.map (eval_string_exp true venv_args pos) args in
         let venv = venv_add_args venv pos loc env params args in
            eval_body_value venv pos body
    | ValPrim (_, be_eager, f) ->
         let args = List.map (eval_string_exp be_eager venv_args pos) args in
            venv_apply_prim_fun f venv pos loc args
    | v ->
         if args = [] then
            v
         else
            let print_error buf =
               fprintf buf "@[<v 3>illegal function application:@ @[<hv 3>function:@ %a@]" pp_print_value v;
               List.iter (fun arg ->
                     fprintf buf "@ @[<hv 3>arg = %a@]" pp_print_string_exp arg) args;
               fprintf buf "@]"
            in
               raise (OmakeException (pos, LazyError print_error))

(*
 * Get a function from a value.
 *)
and eval_fun venv pos v =
   match eval_value venv pos v with
      ValFun (arity, env, params, body) ->
         let f venv pos loc args =
            let venv = venv_add_args venv pos loc env params args in
            let _, result = eval_exp venv ValNone body in
               result
         in
            arity, true, f
    | ValFunValue (arity, env, params, body) ->
         let f venv pos loc args =
            let venv = venv_add_args venv pos loc env params args in
               eval_body_value venv pos body
         in
            arity, true, f
    | ValPrim (arity, be_eager, f) ->
         arity, be_eager, venv_apply_prim_fun f
    | ValBody body ->
         let arity = ArityExact 0 in
         let f venv pos loc args =
            if args <> [] then
               raise (OmakeException (loc_pos loc pos, ArityMismatch (arity, List.length args)));
            let _, result = eval_exp venv ValNone body in
               result
         in
            arity, true, f
    | _ ->
         raise (OmakeException (pos, StringError "not a function"))

(*
 * Get an object from a variable.
 *)
and eval_object venv pos x =
   try eval_object_exn venv pos x with
      Not_found ->
         raise (OmakeException (pos, StringError "not an object"))

and eval_object_exn venv pos x =
   match eval_value venv pos x with
      ValObject env ->
         env
    | ValEnv (e, l) ->
         raise Not_found
    | ValInt _ ->
         create_object venv x int_object_sym
    | ValFloat _ ->
         create_object venv x float_object_sym
    | ValData _
    | ValQuote _ ->
         create_object venv x string_object_sym
    | ValSequence _
    | ValString _
    | ValNone ->
         create_object venv x sequence_object_sym
    | ValArray _ ->
         create_object venv x array_object_sym
    | ValFun _
    | ValFunValue _
    | ValPrim _ ->
         create_object venv x fun_object_sym
    | ValRules _ ->
         create_object venv x rule_object_sym
    | ValNode _ ->
         create_object venv x file_object_sym
    | ValDir _ ->
         create_object venv x dir_object_sym
    | ValBody _ ->
         create_object venv x body_object_sym
    | ValChannel (InChannel, _) ->
         create_object venv x in_channel_object_sym
    | ValChannel (OutChannel, _) ->
         create_object venv x out_channel_object_sym
    | ValChannel (InOutChannel, _) ->
         create_object venv x in_out_channel_object_sym
    | ValMap _ ->
         create_object venv x map_object_sym
    | ValOther (ValLexer _) ->
         create_object venv x lexer_object_sym
    | ValOther (ValParser _) ->
         create_object venv x parser_object_sym
    | ValOther (ValLocation _) ->
         create_object venv x location_object_sym
    | ValOther (ValPosition _) ->
         create_object venv x position_object_sym
    | ValClass _ ->
         raise (Invalid_argument "internal error: dereferenced $class")
    | ValCases _ ->
         raise (Invalid_argument "internal error: dereferenced cases")
    | ValApply _
    | ValSuperApply _
    | ValMethodApply _ ->
         raise (Invalid_argument "find_object")

and create_object venv x v =
   let obj = venv_find_var_exn venv ScopePublic v in
      match obj with
         ValObject env ->
            venv_add_field env builtin_sym x
       | _ ->
            raise Not_found

(*
 * Method paths.
 *)
and eval_method_var_exn venv env pos vl =
   match vl with
      [v] ->
         env, venv_find_field_exn env v
    | v :: vl ->
         let v = venv_find_field_exn env v in
         let obj = eval_object_exn venv pos v in
            eval_method_var_exn venv obj pos vl
    | [] ->
         raise (OmakeException (pos, StringError "empty method name"))

and eval_method_var_aux venv envl pos vl =
   match envl with
      [env] ->
         eval_method_var_exn venv env pos vl
    | env :: envl ->
         (try eval_method_var_exn venv env pos vl with
             Not_found ->
                eval_method_var_aux venv envl pos vl)
    | [] ->
         raise Not_found

and eval_method_var venv scope pos loc vl =
   let envl = venv_current_objects venv scope in
      try eval_method_var_aux venv envl pos vl with
         Not_found ->
            let pos = string_pos "eval_method_var" (loc_pos loc pos) in
               raise (OmakeException (pos, UnboundMethod vl))

(*
 * Simplify a quoted value if possible.
 * Strings are concatenated.
 *)
and simplify_quote_val venv pos el =
   match el with
      [ValString s]
    | [ValData s] ->
         ValData s
    | _ ->
         let buf = Buffer.create 32 in
         let flush vl =
            if Buffer.length buf = 0 then
               vl
            else
               let s = Buffer.contents buf in
                  Buffer.clear buf;
                  ValData s :: vl
         in
         let rec collect vl el =
            match el with
               e :: el ->
                  (match eval_value venv pos e with
                      ValString s
                    | ValData s ->
                         Buffer.add_string buf s;
                         collect vl el
                    | v ->
                         collect (v :: flush vl) el)
             | [] ->
                  List.rev (flush vl)
         in
            match collect [] el with
               [ValString s]
             | [ValData s] ->
                  ValData s
             | el ->
                  ValQuote el

(*
 * Evaluate a string expression.
 *)
and eval_string_exp be_eager venv pos s =
   match s with
      ConstString (_, s) ->
         ValString s
    | ApplyString (loc, strategy, scope, v, []) ->
         if strategy_is_eager be_eager strategy then
            eval_var venv pos loc (venv_find_var venv scope pos loc v)
         else
            ValApply (loc, scope, v, [])
    | ApplyString (loc, strategy, scope, v, args) ->
         if strategy_is_eager be_eager strategy then
            eval_apply_string_exp venv venv pos loc (venv_find_var venv scope pos loc v) args
         else
            let args = List.map (eval_string_exp false venv pos) args in
               ValApply (loc, scope, v, args)
    | SuperApplyString (loc, strategy, scope, super, v, args) ->
         if strategy_is_eager be_eager strategy then
            let obj = venv_find_super venv pos loc super in
            let v = venv_find_field obj pos v in
               eval_apply_string_exp venv venv pos loc v args
         else
            let args = List.map (eval_string_exp false venv pos) args in
               ValSuperApply (loc, scope, super, v, args)
    | MethodApplyString (loc, strategy, scope, vl, args) ->
         if strategy_is_eager be_eager strategy then
            let obj, v = eval_method_var venv scope pos loc vl in
            let venv_obj = venv_with_object venv obj in
               eval_apply_string_exp venv venv_obj pos loc v args
         else
            let args = List.map (eval_string_exp false venv pos) args in
               ValMethodApply (loc, scope, vl, args)
    | SequenceString (loc, sl) ->
         ValSequence (List.map (eval_string_exp be_eager venv pos) sl)
    | BodyString (_, e) ->
         ValBody e
    | ArrayString (_, el) ->
         ValArray (List.map (eval_string_exp be_eager venv pos) el)
    | ArrayOfString (_, e) ->
         let v = eval_string_exp be_eager venv pos e in
            ValArray (values_of_value venv pos v)
    | ExpString (_, e) ->
         let _, result = eval_exp venv ValNone e in
            result
    | CasesString (_, cases) ->
         let cases =
            List.map (fun (v, e1, e2) ->
                  v, eval_string_exp be_eager venv pos e1, ValBody e2) cases
         in
            ValCases cases
    | QuoteString (_, el) ->
         simplify_quote_val venv pos (List.map (eval_string_exp be_eager venv pos) el)
    | ThisString (_, scope) ->
         ValObject (venv_this_object venv scope)

(*
 * Short forms of the string evaluator.
 *)
and lazy_string_exp venv pos s =
   eval_string_exp false venv pos s

and eager_string_exp venv pos s =
   eval_string_exp true venv pos s

(*
 * Evaluate an expression.
 *)
and eval_exp venv result e =
   let pos = string_pos "eval_exp" (ir_exp_pos e) in
   let venv, result =
      match e with
         LetVarExp (_, scope, v, flag, s) ->
            eval_let_var_exp venv pos scope v flag s
       | LetFunExp (loc, scope, v, params, body) ->
            eval_let_fun_exp venv pos loc scope v params body
       | LetObjectExp (_, scope, v, e) ->
            eval_let_object_exp venv pos scope v e
       | LetThisExp (_, e) ->
            eval_let_this_exp venv pos e
       | ShellExp (loc, e) ->
            eval_shell_exp venv pos loc e
       | IfExp (_, cases) ->
            eval_if_exp venv pos cases
       | SequenceExp (_, e) ->
            eval_sequence_exp venv pos e
       | SectionExp (_, _, e) ->
            eval_section_exp venv pos e
       | IncludeExp (loc, s, e) ->
            eval_include_exp venv pos loc s e
       | ApplyExp (loc, scope, f, args) ->
            eval_apply_exp venv scope pos loc f args
       | SuperApplyExp (loc, scope, super, v, args) ->
            eval_super_apply_exp venv scope pos loc super v args
       | MethodApplyExp (loc, scope, vl, args) ->
            eval_method_apply_exp venv scope pos loc vl args
       | ReturnExp (_, s) ->
            eval_return_exp venv pos s
       | ReturnSaveExp _ ->
            eval_return_save_exp venv pos
       | ReturnObjectExp (_, names) ->
            eval_return_object_exp venv pos names
   in
      add_exports venv pos result

(*
 * Variable definitions.
 *)
and eval_let_var_exp venv pos scope v flag s =
   let pos = string_pos "eval_var_exp" pos in
   let s = eager_string_exp venv pos s in
   let s =
      match flag with
         VarDefNormal ->
            s
       | VarDefAppend ->
            let v = venv_get_var venv scope pos v in
               if is_array_value v || is_array_value s then
                  let vl = values_of_value venv pos v in
                  let sl = values_of_value venv pos s in
                     ValArray (vl @ sl)
               else
                  ValSequence [v; ValString " "; s]
   in
   let venv = venv_add_var venv scope pos v s in
      venv, s

(*
 * Function definitions.
 *)
and eval_let_fun_exp venv pos loc scope v params body =
   let pos = string_pos "eval_fun_exp" pos in
   let env = venv_get_env venv in
   let e = ValFun (ArityExact (List.length params), env, params, body) in
   let venv = venv_add_var venv scope pos v e in
      venv, e

(*
 * Shell expression.
 *)
and eval_shell_exp venv pos loc e =
   let pos = string_pos "eval_shell_exp" pos in
   let v = venv_find_var venv ScopePublic pos loc system_sym in
   let s = lazy_string_exp venv pos e in
   let result = eval_apply venv pos loc v [s] in
      venv, result

(*
 * Conditionals.
 * The test should expand to a Boolean of some form.
 *)
and eval_if_cases venv pos cases =
   match cases with
      (s, e) :: cases ->
         let s = eager_string_exp venv pos s in
         let b = bool_of_value venv pos s in
            if b then
               let _, result = eval_exp venv ValNone e in
                  result
            else
               eval_if_cases venv pos cases
    | [] ->
         ValNone

and eval_if_exp venv pos cases =
   let pos = string_pos "eval_if_exp" pos in
   let result = eval_if_cases venv pos cases in
      venv, result

(*
 * Sequence.
 *)
and eval_sequence venv pos result el =
   match el with
      [e] ->
         eval_exp venv result e
    | e :: el ->
         let venv, result = eval_exp venv result e in
            eval_sequence venv pos result el
    | [] ->
         venv, result

and eval_sequence_exp venv pos el =
   let pos = string_pos "eval_sequence_exp" pos in
      eval_sequence venv pos ValNone el

and eval_section_exp venv pos el =
   let pos = string_pos "eval_section_exp" pos in
   let _, result = eval_sequence venv pos ValNone el in
      venv, result

(*
 * Object.
 * The argument string is ignored.
 * Push a new object.
 *)
and eval_let_object_exp venv pos scope v el =
   let pos = string_pos "eval_let_object_exp" pos in
   let _, result = eval_sequence (venv_define_object venv) pos ValNone el in
   let venv = venv_add_var venv scope pos v result in
      venv, result

(*
 * This.
 * Set the current object to the given object.
 *)
and eval_let_this_exp venv pos s =
   let pos = string_pos "eval_this_exp" pos in
   let obj = eager_string_exp venv pos s in
   let obj = eval_object venv pos obj in
   let venv = venv_with_object venv obj in
      venv, ValObject obj

(*
 * Include a file.
 * The environment after the file is evaluated is used in the rest
 * of this file.
 *)
and eval_include_exp venv pos loc s commands =
   let pos = string_pos "eval_include" pos in
   let node = eager_string_exp venv pos s in
   let node = file_of_value venv pos node in
   let venv = venv_add_file venv node in
   let venv = include_file venv pos loc node in
      venv, ValNone

(*
 * Function application.
 *)
and eval_apply_exp venv scope pos loc f args =
   let pos = string_pos "eval_apply_exp" pos in
   let result = eval_apply_string_exp venv venv pos loc (venv_find_var venv scope pos loc f) args in
      venv, result

and eval_super_apply_exp venv scope pos loc super v args =
   let pos = string_pos "eval_super_apply_exp" pos in
   let obj = venv_find_super venv pos loc super in
   let v = venv_find_field obj pos v in
   let result = eval_apply_string_exp venv venv pos loc v args in
      venv, result

and eval_method_apply_exp venv scope pos loc vl args =
   let pos = string_pos "eval_method_apply_exp" pos in
   let obj, v = eval_method_var venv scope pos loc vl in
   let venv' = venv_with_object venv obj in
   let result = eval_apply_string_exp venv venv' pos loc v args in
      venv, result

(*
 * Return a value.  This is always the last statement.
 *)
and eval_return_exp venv pos s =
   let pos = string_pos "eval_return_exp" pos in
   let result = eager_string_exp venv pos s in
      venv, result

and eval_return_save_exp venv pos =
   let pos = string_pos "eval_return_save_exp" pos in
   let result = ValEnv (venv, ExportFile) in
      venv, result

and eval_return_object_exp venv pos names =
   let result = venv_current_object venv names in
      venv, ValObject result

(*
 * Include a file.
 *)
and eval_include_file venv pos loc node =
   let parse_ir node =
      let result = parse_ir node in
         NodeSet.singleton node, result
   in
   let _, ir = compile_file include_name parse_ir Omake_magic.ir_magic venv pos loc node in
   let _, result = eval_exp venv ValNone ir in
      result

and include_file venv pos loc target =
   let pos = string_pos "eval_include_file" pos in
   let result = eval_include_file venv pos loc target in
      fst (add_include venv pos result)

(*
 * Parse and evaluate a file as if it were an object.
 *)
and eval_object_file venv pos loc node =
   let parse_ir node =
      let result = parse_ir node in
         NodeSet.singleton node, result
   in
   let parse_obj node =
      let senv, ir = compile_file include_name parse_ir Omake_magic.ir_magic venv pos loc node in
      let venv = venv_get_pervasives venv node in
      let venv = venv_define_object venv in
      let venv, result = eval_exp venv ValNone ir in
      let names = senv_class_names senv in
      let obj = venv_current_object venv names in
      let files = venv_files venv in
         files, obj
   in
      compile_file object_name parse_obj Omake_magic.obj_magic venv pos loc node

(************************************************************************
 * Evaluator.
 *)
and eval venv e =
   let _, result = eval_exp venv ValNone e in
      result

(************************************************************************
 * Project compiler.
 *)
let compile venv =
   let node = venv_intern venv makeroot_name in
   let venv = venv_add_file venv node in
   let loc = bogus_loc (Node.fullname node) in
   let pos = string_pos "compile" (loc_exp_pos loc) in
   let _ = eval_include_file venv pos loc node in
      if debug print_rules then
         eprintf "@[<hv 3>Rules:%a@]@." pp_print_explicit_rules venv

(************************************************************************
 * Dependencies.
 *)
let compile_deps venv buf =
   let deps = Omake_ast_lex.parse_deps buf in
   let venv_empty = SymbolTable.empty in
      List.map (fun (target, source, loc) ->
            let pos = string_pos "compile_deps" (loc_exp_pos loc) in
            let target = build_string senv_empty target pos in
            let source = build_string senv_empty source pos in
            let target = eager_string_exp venv pos target in
            let source = eager_string_exp venv pos source in
            let targets = strings_of_value venv pos target in
            let sources = strings_of_value venv pos source in
               targets, sources) deps

(*!
 * @docoff
 *
 * -*-
 * Local Variables:
 * Caml-master: "compile"
 * End:
 * -*-
 *)
